# frozen_string_literal: true

require 'pathname'
require_relative 'pathtree/version'

class Pathtree < Pathname
  module StringRefinements
    refine String do
      def prefix_dot
        return self if start_with?('.')

        ".#{self}"
      end

      # Split at last underscore `_` and turn last part into file extension
      #
      #   "aaa_bbb_ccc".split_extension #=> aaa_bbb.ccc
      def split_extension
        sub(/\A[a-z0-9_]*\K_(?=[a-z0-9]+\Z)/, '.')
      end
    end
  end

  module ObjectRefinements
    using StringRefinements

    refine Object do
      # Use Pathtree::StringRefinements#split_extension by default
      def file_path(name:)
        path(name: name, proc_for_nil: -> { _1.split_extension })
      end

      def directory_path(name:)
        path(name: name, proc_for_nil: -> { _1 })
      end

      private

      def path(name:, proc_for_nil:) # :nodoc:
        case self
        in String
          self
        in NilClass
          proc_for_nil.call(name.to_s)
        in Proc
          call(name.to_s)
        end
      end
    end
  end

  class Dsl
    def self.read(file_path, working_directory: Pathname.pwd)
      new(working_directory: working_directory)
        .tap { _1.send(:read, file_path) }
        .instance_variable_get(:@working_directory)
    end

    # See also Pathtree::Dsl.read.
    def initialize(working_directory: Pathname.pwd)
      @working_directory = Pathtree.new(working_directory)
    end

    using ObjectRefinements

    def file(name, path = nil)
      path = path.file_path(name: name)

      # NOTE: Temporal variable is needed
      # since define_singleton_method's block is evaluated using instance_eval
      # which is same as define_method.
      working_directory = @working_directory

      @working_directory.define_singleton_method(name) { working_directory / path }
    end

    def directory(name, path = nil, &block)
      path = path.directory_path(name: name)

      # NOTE: Temporal variable is needed
      # since define_singleton_method's block is evaluated using instance_eval
      # which is same as define_method.
      working_directory = @working_directory / path

      @working_directory.define_singleton_method(name) do
        Dsl.new(working_directory: working_directory)
           .tap { _1.instance_eval(&block) if block_given? }
           .instance_variable_get(:@working_directory)
      end
    end

    alias dir directory

    private

    using StringRefinements

    def read(file_path)
      instance_eval File.read(file_path)
    end
  end
end
